/* -*- linux-c -*- */

/* gtkcombobox.c - GtkComboBox widget implementation
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "gtkcombobox.h"

#include <gtk/gtkcontainer.h>
#include <gtk/gtkmain.h>
#include <gtk/gtkarrow.h>
#include <gtk/gtktogglebutton.h>
#include <gtk/gtkwindow.h>
#include <gtk/gtktearoffmenuitem.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkeventbox.h>
#include <gtk/gtkvbox.h>
#include <gtk/gtkframe.h>
#include <gdk/gdkkeysyms.h>

enum {
	COMBO_ACTIVATE,
	COMBO_POPUP,
	COMBO_POPDOWN,
	LAST_SIGNAL
};

static void     gtk_combo_box_draw             (GtkWidget *widget,
						GdkRectangle *area);
static gint     gtk_combo_box_expose           (GtkWidget *widget,
						GdkEventExpose *event);
static void     gtk_combo_box_map              (GtkWidget *widget);
static void     gtk_combo_box_unmap            (GtkWidget *widget);
static void     gtk_combo_box_size_request     (GtkWidget *widget,
					        GtkRequisition *requisition);
static void     gtk_combo_box_size_allocate    (GtkWidget *widget,
						GtkAllocation *allocation);

static void     gtk_combo_box_finalize         (GObject *object);


static void     gtk_combo_box_forall          (GtkContainer *container,
					       gboolean      include_internals,
					       GtkCallback   callback,
					       gpointer      callback_data);
static void     gtk_combo_box_remove           (GtkContainer *container,
						GtkWidget    *widget);

static void     gtk_combo_box_class_init       (GtkComboBoxClass *class);
static void     gtk_combo_box_init             (GtkComboBox *cbox);

static gint     gtk_combo_box_button_press     (GtkWidget *widget,
						GdkEventButton *event,
						GtkComboBox *cbox);
static void     gtk_combo_box_deactivate_arrow (GtkComboBox *cbox);
static void     gtk_combo_box_get_pos          (GtkComboBox *cbox, 
						int *x, int *y);
static gint     gtk_combo_box_key_press        (GtkWidget *widget,
						GdkEventKey *event,
						GtkComboBox *cbox);
static gboolean gtk_combo_box_popup_delete     (GtkWidget *widget,
						GdkEventAny *event,
						GtkComboBox *cbox);
static void     gtk_combo_box_popup_display    (GtkComboBox *cbox);
static void     gtk_combo_box_popup_hide_unconditional (GtkComboBox *cbox);
static void     gtk_combo_box_popup_reparent   (GtkWidget *popup,
						GtkWidget *new_parent,
						gboolean   unrealize);
static void     gtk_combo_box_popup_tear_off   (GtkComboBox *cbox,
						gboolean set_position);
static void     gtk_combo_box_state_change     (GtkWidget *widget,
						GtkStateType old_state,
						GtkComboBox *cbox);
static void     gtk_combo_box_set_tearoff_state (GtkComboBox *cbox,
						 gboolean torn_off);
static void     gtk_combo_box_tearoff_bg_copy  (GtkComboBox *cbox);
static gboolean gtk_combo_box_tearable_button_release (GtkWidget *widget,
						       GdkEventButton *event,
						       GtkComboBox *cbox);
static gboolean gtk_combo_box_tearable_enter_leave (GtkWidget *w,
						    GdkEventCrossing *event,
						    gpointer data);
static int      gtk_combo_box_toggle_pressed   (GtkToggleButton *tbutton,
						GtkComboBox *cbox);

static GtkContainerClass *gtk_combo_box_parent_class = NULL;
static gint gtk_combo_box_signals[LAST_SIGNAL] = {0};

/* widget class functions */

static void
gtk_combo_box_draw (GtkWidget *widget, GdkRectangle *area)
{
	GtkComboBox      *cbox;
	GdkRectangle      child_area;

	g_return_if_fail (widget != NULL);
	g_return_if_fail (GTK_IS_COMBO_BOX (widget));

	if (GTK_WIDGET_DRAWABLE (widget)) {
		cbox = GTK_COMBO_BOX (widget);
		
		if (gtk_widget_intersect (cbox->display_widget, area,
					  &child_area))
			gtk_widget_draw (cbox->display_widget, &child_area);
		
		if (gtk_widget_intersect (cbox->arrow_button, area,
					  &child_area))
			gtk_widget_draw (cbox->arrow_button, &child_area);
	}
}

static gint
gtk_combo_box_expose (GtkWidget *widget, GdkEventExpose *event)
{
	GtkComboBox   *cbox;
	GdkEventExpose child_event;

	g_return_val_if_fail (widget != NULL, FALSE);
	g_return_val_if_fail (GTK_IS_COMBO_BOX (widget), FALSE);
	g_return_val_if_fail (event != NULL, FALSE);

	if (GTK_WIDGET_DRAWABLE (widget)) {
		cbox = GTK_COMBO_BOX (widget);

		child_event = *event;

		if (gtk_widget_intersect (cbox->display_widget,
					  &event->area,
					  &child_event.area))
			gtk_widget_event (cbox->display_widget,
					  (GdkEvent *)&child_event);

		child_event = *event;
		
		if (gtk_widget_intersect (cbox->arrow_button,
					  &event->area,
					  &child_event.area))
			gtk_widget_event (cbox->arrow_button,
					  (GdkEvent *)&child_event);
	}
	
	return FALSE;
}

static void
gtk_combo_box_map (GtkWidget *widget)
{
	GtkComboBox    *cbox;
	
	g_return_if_fail (widget != NULL);
	g_return_if_fail (GTK_IS_COMBO_BOX (widget));

	cbox = GTK_COMBO_BOX (widget);
	GTK_WIDGET_SET_FLAGS (widget, GTK_MAPPED);

	if (GTK_WIDGET_VISIBLE (cbox->display_widget)
	    && !GTK_WIDGET_MAPPED (cbox->display_widget))
		gtk_widget_map (cbox->display_widget);
	
	if (GTK_WIDGET_VISIBLE (cbox->arrow_button)
	    && !GTK_WIDGET_MAPPED (cbox->arrow_button))
		gtk_widget_map (cbox->arrow_button);
}

static void
gtk_combo_box_unmap (GtkWidget *widget)
{
	GtkComboBox   *cbox;
	
	g_return_if_fail (widget != NULL);
	g_return_if_fail (GTK_IS_COMBO_BOX (widget));
	
	cbox = GTK_COMBO_BOX (widget);
	GTK_WIDGET_UNSET_FLAGS (widget, GTK_MAPPED);

	if (GTK_WIDGET_VISIBLE (cbox->display_widget)
	    && GTK_WIDGET_MAPPED (cbox->display_widget))
		gtk_widget_unmap (cbox->display_widget);
	
	if (GTK_WIDGET_VISIBLE (cbox->arrow_button)
	    && GTK_WIDGET_MAPPED (cbox->arrow_button))
		gtk_widget_unmap (cbox->arrow_button);
}

static void
gtk_combo_box_size_request (GtkWidget *widget, GtkRequisition *requisition)
{
	GtkComboBox    *cbox;
	GtkRequisition  child_requisition;

	g_return_if_fail (widget != NULL);
	g_return_if_fail (GTK_IS_COMBO_BOX (widget));
	g_return_if_fail (requisition != NULL);

	cbox = GTK_COMBO_BOX (widget);

	requisition->width = GTK_CONTAINER (cbox)->border_width * 2;
	requisition->height = GTK_CONTAINER (cbox)->border_width * 2;
	
	if (cbox->display_widget) {
		gtk_widget_size_request (cbox->display_widget,
					 &child_requisition);
		
		requisition->width += child_requisition.width;
		requisition->height += child_requisition.height;
	}

	gtk_widget_size_request (cbox->arrow_button,
				 &child_requisition);
	requisition->width += child_requisition.width;
	requisition->height += child_requisition.height;
}

static void
gtk_combo_box_size_allocate (GtkWidget *widget, GtkAllocation *allocation)
{
	GtkComboBox     *cbox;
	GtkAllocation    alloc;
	GtkRequisition   child_requisition;
	gint             border_width;

	g_return_if_fail (widget != NULL);
	g_return_if_fail (GTK_IS_COMBO_BOX (widget));
	g_return_if_fail (allocation != NULL);
	
	cbox = GTK_COMBO_BOX (widget);
	widget->allocation = *allocation;
	
	border_width = GTK_CONTAINER (cbox)->border_width;

	alloc.x = allocation->x + border_width;
	
	if (cbox->display_widget 
	    && GTK_WIDGET_VISIBLE (cbox->display_widget)) {
		gtk_widget_get_child_requisition (cbox->display_widget,
						  &child_requisition);
		
		alloc.width = child_requisition.width;
		alloc.height = child_requisition.height;

		alloc.y = allocation->y + 
			(allocation->height - child_requisition.height) / 2;
		
		gtk_widget_size_allocate (cbox->display_widget, &alloc);
		
		alloc.x += child_requisition.width;
	}

	if (GTK_WIDGET_VISIBLE (cbox->arrow_button)) {
		gtk_widget_get_child_requisition (cbox->arrow_button,
						  &child_requisition);
		
		alloc.width = child_requisition.width;
		alloc.height = child_requisition.height;

		alloc.y = allocation->y + 
			(allocation->height - child_requisition.height) / 2;
		
		gtk_widget_size_allocate (cbox->arrow_button, &alloc);
		
		alloc.x += child_requisition.width;
	}
}

/* object class functions */

static void
gtk_combo_box_finalize (GObject *object)
{
	GtkComboBox *cbox;
	
	g_return_if_fail (object != NULL);
	g_return_if_fail (GTK_IS_COMBO_BOX (object));

	cbox = GTK_COMBO_BOX (object);

	if (cbox->tearoff_window) {
		gtk_object_unref (GTK_OBJECT (cbox->tearoff_window));
		cbox->tearoff_window = NULL;
	}

	if (cbox->toplevel && GTK_IS_WIDGET (cbox->toplevel)) {
		gtk_object_unref (GTK_OBJECT (cbox->toplevel));
		cbox->toplevel = NULL;
	}

	G_OBJECT_CLASS (gtk_combo_box_parent_class)->finalize (object);
}

/* container class functions */

static void
gtk_combo_box_forall (GtkContainer *container,
		      gboolean      include_internals,
		      GtkCallback   callback,
		      gpointer      callback_data)
{
	GtkComboBox  *cbox;
	
	g_return_if_fail (container != NULL);
	g_return_if_fail (GTK_IS_COMBO_BOX (container));
	g_return_if_fail (callback != NULL);
	
	cbox = GTK_COMBO_BOX (container);
	
	if (cbox->display_widget)
		(*callback) (cbox->display_widget, callback_data);
	
	if (cbox->arrow_button)
		(*callback) (cbox->arrow_button, callback_data);
}

static void
gtk_combo_box_remove (GtkContainer *container, GtkWidget *widget)
{
	gboolean       was_visible;
	GtkComboBox   *cbox;

	g_return_if_fail (container != NULL);
	g_return_if_fail (GTK_IS_COMBO_BOX (container));
	g_return_if_fail (widget != NULL);
	g_return_if_fail (GTK_IS_WIDGET (widget));

	cbox = GTK_COMBO_BOX (container);
	
	/**
	 * this is a very limited widget :) Only able to remove
	 * the display_widget.
	 */

	if (widget != cbox->display_widget) 
		return;

	was_visible = GTK_WIDGET_VISIBLE (widget);
	gtk_widget_unparent (widget);
	
	if (was_visible && GTK_WIDGET_VISIBLE (container))
		gtk_widget_queue_resize (GTK_WIDGET (container));
}

/* class functions */

static void
gtk_combo_box_class_init (GtkComboBoxClass *class)
{
	GObjectClass      *gobject_class = G_OBJECT_CLASS (class);
	GtkObjectClass    *object_class;
	GtkWidgetClass    *widget_class;
	GtkContainerClass *container_class;

	object_class = (GtkObjectClass *)class;
	widget_class = (GtkWidgetClass *)class;
	container_class = (GtkContainerClass *)class;
	gtk_combo_box_parent_class = 
		g_type_class_peek_parent (class);

	gobject_class->finalize = gtk_combo_box_finalize;

	widget_class->map = gtk_combo_box_map;
	widget_class->unmap = gtk_combo_box_unmap;
	widget_class->draw = gtk_combo_box_draw;
	widget_class->expose_event = gtk_combo_box_expose;
	widget_class->size_request = gtk_combo_box_size_request;
	widget_class->size_allocate = gtk_combo_box_size_allocate;

	container_class->forall = gtk_combo_box_forall;
	container_class->remove = gtk_combo_box_remove;
	
	gtk_combo_box_signals[COMBO_ACTIVATE] = gtk_signal_new (
		"activate",
		GTK_RUN_LAST,
		gtk_combo_box_get_type (),
		GTK_SIGNAL_OFFSET (GtkComboBoxClass, activate),
		gtk_marshal_NONE__NONE,
		GTK_TYPE_NONE, 0);
	gtk_combo_box_signals[COMBO_POPUP] = gtk_signal_new ( 
		"popup",
		GTK_RUN_LAST,
		gtk_combo_box_get_type (),
		GTK_SIGNAL_OFFSET (GtkComboBoxClass, popup),
		gtk_marshal_NONE__NONE,
		GTK_TYPE_NONE, 0);
	gtk_combo_box_signals[COMBO_POPDOWN] = gtk_signal_new ( 
		"popdown",
		GTK_RUN_LAST,
		gtk_combo_box_get_type (),
		GTK_SIGNAL_OFFSET (GtkComboBoxClass, popdown),
		gtk_marshal_NONE__NONE,
		GTK_TYPE_NONE, 0);
	
	gtk_object_class_add_signals (object_class,
				      gtk_combo_box_signals,
				      LAST_SIGNAL);
}

static void
gtk_combo_box_init (GtkComboBox *cbox)
{
	GtkWidget  *arrow;
	GdkCursor  *cursor;

	GTK_WIDGET_SET_FLAGS (cbox, GTK_NO_WINDOW);
	GTK_WIDGET_UNSET_FLAGS (cbox, GTK_CAN_FOCUS);

	cbox->arrow_button = gtk_toggle_button_new ();
	GTK_WIDGET_UNSET_FLAGS (cbox->arrow_button, GTK_CAN_FOCUS);
	arrow = gtk_arrow_new (GTK_ARROW_DOWN, GTK_SHADOW_IN);
	gtk_container_add (GTK_CONTAINER (cbox->arrow_button),
			   arrow);
	gtk_signal_connect (GTK_OBJECT (cbox->arrow_button), "toggled",
			    GTK_SIGNAL_FUNC (gtk_combo_box_toggle_pressed),
			    cbox);
	gtk_signal_connect (GTK_OBJECT (cbox->arrow_button),
			    "state-changed",
			    GTK_SIGNAL_FUNC (gtk_combo_box_state_change),
			    cbox );
	gtk_widget_set_parent (cbox->arrow_button, GTK_WIDGET (cbox));
	gtk_widget_show_all (cbox->arrow_button);
	
	if (GTK_WIDGET_REALIZED (cbox))
		gtk_widget_realize (cbox->arrow_button);
	
	if (GTK_WIDGET_VISIBLE (cbox) 
	     && GTK_WIDGET_VISIBLE (cbox->arrow_button)) {

		if (GTK_WIDGET_MAPPED (cbox))
			gtk_widget_map (cbox->arrow_button);

		gtk_widget_queue_resize (cbox->arrow_button);
	}
	
	cbox->toplevel = gtk_window_new (GTK_WINDOW_POPUP);
	gtk_widget_ref (cbox->toplevel);
	gtk_object_sink (GTK_OBJECT (cbox->toplevel));
	gtk_window_set_policy (GTK_WINDOW (cbox->toplevel),
			       FALSE, TRUE, FALSE);
	gtk_signal_connect (GTK_OBJECT (cbox->toplevel),
			    "button_press_event",
			    GTK_SIGNAL_FUNC (gtk_combo_box_button_press),
			    cbox);
	gtk_signal_connect (GTK_OBJECT (cbox->toplevel),
			    "key_press_event",
			    GTK_SIGNAL_FUNC (gtk_combo_box_key_press),
			    cbox);
	
	cbox->popup = gtk_event_box_new ();
	gtk_container_add (GTK_CONTAINER (cbox->toplevel),
			   cbox->popup);
	gtk_widget_show (cbox->popup);
	
	gtk_widget_realize (cbox->popup);
	cursor = gdk_cursor_new (GDK_TOP_LEFT_ARROW);
	gdk_window_set_cursor (cbox->popup->window, cursor);
	gdk_cursor_destroy (cursor);
	
	cbox->frame = gtk_frame_new (NULL);
	gtk_frame_set_shadow_type (GTK_FRAME (cbox->frame),
				   GTK_SHADOW_OUT);
	gtk_container_add (GTK_CONTAINER (cbox->popup),
			   cbox->frame);
	gtk_widget_show (cbox->frame);
	
	cbox->torn_off = FALSE;
	cbox->tearoff_window = NULL;
}

/* utilily functions */

static gint
gtk_combo_box_button_press (GtkWidget *widget, GdkEventButton *event,
			    GtkComboBox *cbox)
{
	GtkWidget    *child;

	child = gtk_get_event_widget ((GdkEvent *)event);
	if (child != widget) {
		while (child) {
			if (child == widget)
				return FALSE;
			child = child->parent;
		}
	}

	gtk_combo_box_popup_hide (cbox);

	return TRUE;
}

static void
gtk_combo_box_deactivate_arrow (GtkComboBox *cbox)
{
	GtkToggleButton   *arrow;
	
	arrow = GTK_TOGGLE_BUTTON (cbox->arrow_button);
	gtk_signal_handler_block_by_func ( 
		GTK_OBJECT (arrow),
		GTK_SIGNAL_FUNC (gtk_combo_box_toggle_pressed), cbox);
	
	gtk_toggle_button_set_active (arrow, FALSE);
	
	gtk_signal_handler_unblock_by_func (
		GTK_OBJECT (arrow),
		GTK_SIGNAL_FUNC (gtk_combo_box_toggle_pressed), cbox);
}

static void
gtk_combo_box_get_pos (GtkComboBox *cbox, int *x, int *y)
{
	int        ph, pw;
	GtkWidget *wcombo = GTK_WIDGET (cbox);

	gdk_window_get_origin (wcombo->window, x, y);
	*y += wcombo->allocation.height + wcombo->allocation.y;
	*x += wcombo->allocation.x;

	ph = cbox->popup->allocation.height;
	pw = cbox->popup->allocation.width;

	if ((*y + ph) > gdk_screen_height ())
		*y = gdk_screen_height () - ph;

	if ((*x + pw) > gdk_screen_width ())
		*x = gdk_screen_width () - pw;
}

static gint
gtk_combo_box_key_press (GtkWidget *widget, GdkEventKey *event,
			 GtkComboBox *cbox)
{
	if (event->keyval == GDK_Escape) {
		gtk_combo_box_popup_hide_unconditional (cbox);
		return TRUE;
	} else
		return FALSE;
}

static gboolean
gtk_combo_box_popup_delete (GtkWidget *widget, GdkEventAny *event,
			    GtkComboBox *cbox)
{
	gtk_combo_box_popup_hide_unconditional (cbox);
	
	return TRUE;
}

static void
gtk_combo_box_popup_display (GtkComboBox *cbox)
{
	int x, y;

	g_return_if_fail (cbox != NULL);
	g_return_if_fail (GTK_IS_COMBO_BOX (cbox));
	
	if (!cbox->pop_down_widget) {
		/* there isn't a popdown widget, so we won't popdown */

		gtk_combo_box_deactivate_arrow (cbox);

		return;
	}

	if (cbox->torn_off) {
		gtk_combo_box_tearoff_bg_copy (cbox);

		gtk_combo_box_popup_reparent (cbox->popup,
					      cbox->toplevel, TRUE);
	}

	gtk_combo_box_get_pos (cbox, &x, &y);
	
	gtk_widget_set_uposition (cbox->toplevel, x, y);
	gtk_widget_realize (cbox->popup);
	gtk_widget_show (cbox->popup);
	gtk_widget_realize (cbox->toplevel);
	gtk_widget_show (cbox->toplevel);
	
	gtk_grab_add (cbox->toplevel);
	gdk_pointer_grab (cbox->toplevel->window, TRUE,
			  GDK_BUTTON_PRESS_MASK | 
			  GDK_BUTTON_RELEASE_MASK |
			  GDK_POINTER_MOTION_MASK,
			  NULL, NULL, GDK_CURRENT_TIME);

	gtk_signal_emit (GTK_OBJECT (cbox),
			  gtk_combo_box_signals[COMBO_POPUP],
			  NULL);
}

static void
gtk_combo_box_popup_hide_unconditional (GtkComboBox *cbox)
{
	g_return_if_fail (cbox != NULL);
	g_return_if_fail (GTK_IS_COMBO_BOX (cbox));
	
	gtk_widget_hide (cbox->toplevel);
	gtk_widget_hide (cbox->popup);

	if (cbox->torn_off) {
		GTK_TEAROFF_MENU_ITEM (cbox->tearable)->torn_off
			= FALSE;
		gtk_combo_box_set_tearoff_state (cbox, FALSE);
	}
	
	gtk_grab_remove (cbox->toplevel);
	gdk_pointer_ungrab (GDK_CURRENT_TIME);

	gtk_signal_emit (GTK_OBJECT (cbox),
			 gtk_combo_box_signals[COMBO_POPDOWN],
			 NULL);
	
	gtk_combo_box_deactivate_arrow (cbox);
}

static void
gtk_combo_box_popup_reparent (GtkWidget *popup,
			      GtkWidget *new_parent,
			      gboolean   unrealize)
{
	GtkObject   *object = GTK_OBJECT (popup);
	gboolean     was_floating = GTK_OBJECT_FLOATING (object);
	
	gtk_object_ref (object);
	gtk_object_sink (object);
	
	if (unrealize) {
		gtk_object_ref (object);
		gtk_container_remove (GTK_CONTAINER (popup->parent),
				      popup);
		gtk_container_add (GTK_CONTAINER (new_parent), popup);
		gtk_object_unref (object);
	} else
		gtk_widget_reparent (GTK_WIDGET (popup), new_parent);
	
	gtk_widget_set_usize (new_parent, -1, -1);
	
	if (was_floating)
		GTK_OBJECT_SET_FLAGS (object, GTK_FLOATING);
	else
		gtk_object_unref (object);
}

static void
gtk_combo_box_popup_tear_off (GtkComboBox *cbox, gboolean set_position)
{
	int x, y;

	if (!cbox->tearoff_window) {
		GtkWidget *tearoff;
		gchar *title;
		
		tearoff = gtk_window_new (GTK_WINDOW_DIALOG);
		gtk_widget_ref (tearoff);
		gtk_object_sink (GTK_OBJECT (tearoff));

		cbox->tearoff_window = tearoff;
		
		gtk_widget_set_app_paintable (tearoff, TRUE);
		gtk_signal_connect ( 
			GTK_OBJECT (tearoff),
			"key_press_event",
			GTK_SIGNAL_FUNC (gtk_combo_box_key_press),
			GTK_OBJECT (cbox));
		gtk_widget_realize (tearoff);
		title = gtk_object_get_data (GTK_OBJECT (cbox),
					     "gtk-combo-title");
		if (title)
			gdk_window_set_title (tearoff->window, title);
		gtk_window_set_policy (GTK_WINDOW (tearoff),
				       FALSE, TRUE, FALSE);
		gtk_window_set_transient_for
			(GTK_WINDOW (tearoff),
			 GTK_WINDOW (gtk_widget_get_toplevel 
				     GTK_WIDGET (cbox)));
	}
	
	if (GTK_WIDGET_VISIBLE (cbox->popup)) {
		gtk_widget_hide (cbox->toplevel);
		
		gtk_grab_remove (cbox->toplevel);
		gdk_pointer_ungrab (GDK_CURRENT_TIME);
	}

	gtk_combo_box_popup_reparent (cbox->popup,
				      cbox->tearoff_window, FALSE);
	
	gtk_widget_queue_resize (GTK_WIDGET (cbox->popup));
	
	if (set_position) {
		gtk_combo_box_get_pos (cbox, &x, &y);
		gtk_widget_set_uposition (cbox->tearoff_window, x, y);
	}

	gtk_widget_show (GTK_WIDGET (cbox->popup));
	gtk_widget_show (cbox->tearoff_window);
}

static void
gtk_combo_box_state_change (GtkWidget *widget, GtkStateType old_state,
			    GtkComboBox *cbox)
{
	GtkStateType const new_state = GTK_WIDGET_STATE (widget);
	gtk_widget_set_state (cbox->display_widget, new_state);
}

static void
gtk_combo_box_set_tearoff_state (GtkComboBox *cbox,
				 gboolean torn_off)
{
	g_return_if_fail (cbox != NULL);
	g_return_if_fail (GTK_IS_COMBO_BOX (cbox));

	if (cbox->torn_off != torn_off) {
		cbox->torn_off = torn_off;
		
		if (cbox->torn_off) {
			gtk_combo_box_popup_tear_off (cbox, TRUE);
			gtk_combo_box_deactivate_arrow (cbox);
		} else {
			gtk_widget_hide (cbox->tearoff_window);
			gtk_combo_box_popup_reparent (cbox->popup,
						      cbox->toplevel,
						      FALSE);
		}
	}
}

static void
gtk_combo_box_tearoff_bg_copy (GtkComboBox *cbox)
{
	GdkPixmap    *pixmap;
	GdkGC        *gc;
	GdkGCValues   gc_values;
	
	GtkWidget *widget = cbox->popup;

	if (cbox->torn_off) {
		gc_values.subwindow_mode = GDK_INCLUDE_INFERIORS;
		gc = gdk_gc_new_with_values (widget->window,
					     &gc_values,
					     GDK_GC_SUBWINDOW);
		
		pixmap = gdk_pixmap_new (widget->window,
					 widget->allocation.width,
					 widget->allocation.height,
					 -1);
		
		gdk_draw_pixmap (pixmap, gc,
				 widget->window,
				 0, 0, 0, 0, -1, -1);
		gdk_gc_unref (gc);

		gtk_widget_set_usize (cbox->tearoff_window,
				      widget->allocation.width,
				      widget->allocation.height);

		gdk_window_set_back_pixmap (cbox->tearoff_window->window,
					    pixmap, FALSE);
		
		gdk_pixmap_unref (pixmap);
	}
}

static gboolean
gtk_combo_box_tearable_button_release (GtkWidget *widget, 
				       GdkEventButton *event,
				       GtkComboBox *cbox)
{
	GtkTearoffMenuItem  *tearable;
	
	g_return_val_if_fail (widget != NULL, FALSE);
	g_return_val_if_fail (GTK_IS_TEAROFF_MENU_ITEM (widget), FALSE);
	
	tearable = GTK_TEAROFF_MENU_ITEM (widget);
	tearable->torn_off = !tearable->torn_off;
	
	if (!cbox->torn_off) {
		gboolean   need_connect;

		need_connect = (!cbox->tearoff_window);
		gtk_combo_box_set_tearoff_state (cbox, TRUE);
		
		if (need_connect)
			gtk_signal_connect ( 
				GTK_OBJECT (cbox->tearoff_window),
				"delete_event",
				GTK_SIGNAL_FUNC (gtk_combo_box_popup_delete),
				cbox);
	} else
		gtk_combo_box_popup_hide_unconditional (cbox);

	return TRUE;
}

static gboolean
gtk_combo_box_tearable_enter_leave (GtkWidget *w, GdkEventCrossing *event,
				    gpointer data)
{
	gboolean const flag = GPOINTER_TO_INT (data);
	gtk_widget_set_state (w, 
			      flag ? GTK_STATE_PRELIGHT : GTK_STATE_NORMAL);
	
	return FALSE;
}

static int
gtk_combo_box_toggle_pressed (GtkToggleButton *tbutton, GtkComboBox *cbox)
{
	if (tbutton->active)
		gtk_combo_box_popup_display (cbox);
	else
		gtk_combo_box_popup_hide_unconditional (cbox);

	return TRUE;
}

/* implementation of the functions in the header file */

GtkType
gtk_combo_box_get_type (void)
{
	static GtkType   type = 0;

	if (!type) {
		static const GTypeInfo info = {
			sizeof (GtkComboBoxClass),
			NULL, /* base_init */
			NULL, /* base_finalize */
			(GClassInitFunc) gtk_combo_box_class_init,
			NULL, /* class_finalize */
			NULL, /* class_data */
			sizeof (GtkComboBox),
			0, /* n_preallocs */
			(GInstanceInitFunc) gtk_combo_box_init
		};

		type = g_type_register_static (GTK_TYPE_CONTAINER,
					       "GtkComboBox",
					       &info);
	}

	return type;
}

void
gtk_combo_box_construct (GtkComboBox *cbox,
			 GtkWidget   *display_widget,
			 GtkWidget   *pop_down_widget)
{
	GtkWidget    *tearable;
	GtkWidget    *vbox;
	
	g_return_if_fail (cbox != NULL);
	g_return_if_fail (GTK_IS_COMBO_BOX (cbox));
	g_return_if_fail (display_widget != NULL);
	g_return_if_fail (GTK_IS_WIDGET (display_widget));
	
	cbox->pop_down_widget = pop_down_widget;
	cbox->display_widget = NULL;
	
	vbox = gtk_vbox_new (FALSE, 5);
	tearable = gtk_tearoff_menu_item_new ();
	gtk_signal_connect ( 
		GTK_OBJECT (tearable), "enter-notify-event",
		GTK_SIGNAL_FUNC (gtk_combo_box_tearable_enter_leave),
		GINT_TO_POINTER (TRUE));
	gtk_signal_connect (
		GTK_OBJECT (tearable), "leave-notify-event",
		GTK_SIGNAL_FUNC (gtk_combo_box_tearable_enter_leave),
		GINT_TO_POINTER (FALSE));
	gtk_signal_connect ( 
		GTK_OBJECT (tearable), "button-release-event",
		GTK_SIGNAL_FUNC (gtk_combo_box_tearable_button_release),
		(gpointer)cbox);
	gtk_box_pack_start (GTK_BOX (vbox), tearable, FALSE, FALSE, 0);
	if (pop_down_widget)
		gtk_box_pack_start (GTK_BOX (vbox), pop_down_widget,
				    TRUE, TRUE, 0);
	
	cbox->tearable = tearable;
	
	gtk_combo_box_set_display (cbox, display_widget);

	gtk_container_add (GTK_CONTAINER (cbox->frame), vbox);
	gtk_widget_show_all (vbox);
}

GtkWidget *
gtk_combo_box_new (GtkWidget *display_widget,
		   GtkWidget *pop_down_widget)
{
	GtkComboBox *cbox;

	g_return_val_if_fail (display_widget != NULL, NULL);
	g_return_val_if_fail (GTK_IS_WIDGET (display_widget), NULL);
	
	cbox = gtk_type_new (gtk_combo_box_get_type ());
	gtk_combo_box_construct (cbox, display_widget, pop_down_widget);

	return GTK_WIDGET (cbox);
}

/**
 * gtk_combo_box_popup_hide
 * @combo_box: Combo box
 *
 * Hide popup, but not when it is torn off.
 * 
 * This is the external interface - for subclasses and apps which
 * expect a regular combo which doesn't do tearoffs.
 */
void
gtk_combo_box_popup_hide (GtkComboBox *cbox)
{
	g_return_if_fail (cbox != NULL);
	g_return_if_fail (GTK_IS_COMBO_BOX (cbox));

	if (!cbox->torn_off)
		gtk_combo_box_popup_hide_unconditional (cbox);
	else if (GTK_WIDGET_VISIBLE (cbox->toplevel)) {
		gtk_combo_box_popup_tear_off (cbox, FALSE);
		gtk_combo_box_deactivate_arrow (cbox);
	}
}

void
gtk_combo_box_popup_show (GtkComboBox *cbox)
{
	g_return_if_fail (cbox != NULL);
	g_return_if_fail (GTK_IS_COMBO_BOX (cbox));
	
	gtk_combo_box_popup_display (cbox);
}

void
gtk_combo_box_set_arrow_relief (GtkComboBox *cbox,
				GtkReliefStyle relief)
{
	g_return_if_fail (cbox != NULL);
	g_return_if_fail (GTK_IS_COMBO_BOX (cbox));
	
	gtk_button_set_relief (GTK_BUTTON (cbox->arrow_button),
			       relief);
}

/**
 * gtk_combo_box_set_display
 * @cbox: Combo box
 * @display_widget: the widget to be displayed
 *
 * Sets the displayed widget for the @combo_box to be
 * @display_widget.
 */
void
gtk_combo_box_set_display (GtkComboBox *cbox,
			   GtkWidget *display_widget)
{
	g_return_if_fail (cbox != NULL);
	g_return_if_fail (GTK_IS_COMBO_BOX (cbox));
	g_return_if_fail (display_widget != NULL);
	g_return_if_fail (GTK_IS_WIDGET (display_widget));

	/* check if the user isn't setting the same display widget */
	if (cbox->display_widget == display_widget)
		return;

	if (cbox->display_widget)
		gtk_container_remove (GTK_CONTAINER (cbox),
				      cbox->display_widget);
	
	cbox->display_widget = display_widget;
	gtk_widget_ref (cbox->display_widget);
	gtk_widget_set_parent (cbox->display_widget, GTK_WIDGET (cbox));

	if (GTK_WIDGET_REALIZED (cbox))
		gtk_widget_realize (cbox->display_widget);
	
	if (GTK_WIDGET_VISIBLE (cbox) 
	    && GTK_WIDGET_VISIBLE (cbox->display_widget)) {

		if (GTK_WIDGET_MAPPED (cbox))
			gtk_widget_map (cbox->display_widget);

		gtk_widget_queue_resize (cbox->display_widget);
	}
}

void
gtk_combo_box_set_title (GtkComboBox *cbox,
			 const gchar *title)
{
	g_return_if_fail (cbox != NULL);
	g_return_if_fail (GTK_IS_COMBO_BOX (cbox));
	
	gtk_object_set_data_full (GTK_OBJECT (cbox), "gtk-combo-title",
				  g_strdup (title), 
				  (GdkDestroyNotify) g_free);
}
