/* -*- linux-c -*- */

/* gtkcomboboxtext.c - GtkComboBoxText widget implementation
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "gtkcomboboxtext.h"

#include <string.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkentry.h>
#include <gtk/gtkobject.h>
#include <gtk/gtkscrolledwindow.h>
#include <gtk/gtkbindings.h>

enum {
	SELECT,
	AUTOCOMPLETED,
	LAST_SIGNAL
};

static void      gtk_combo_box_text_destroy          (GtkObject *object);

static void      gtk_combo_box_text_class_init   (GtkComboBoxTextClass *class);

static void      gtk_combo_box_text_add_history     (GtkComboBoxText *cboxtext,
						     const gchar     *string);
static GSList   *gtk_combo_box_text_build_item_list  (GtkTreeModel *model);
static void      gtk_combo_box_text_add_to_item_list (GSList       **slist,
						      GtkTreeModel  *model,
						      GtkTreeNode    node);
static void      gtk_combo_box_text_entry_changed (GtkComboBoxText *cboxtext);
static gint      gtk_combo_box_text_entry_key_press (GtkComboBoxText *cboxtext,
						     GdkEventKey     *event,
						     GtkEntry        *entry);
static gchar    *gtk_combo_box_text_intersect       (gint max_len,
						     GSList *strings);
static gboolean  gtk_combo_box_text_match           (GtkComboBoxText *cboxtext,
						     const gchar *string);
static inline gchar *gtk_combo_box_text_node_get_value (GtkTreeModel *model,
							GtkTreeNode   node);
static GtkTreeNode *gtk_combo_box_text_node_new     (GtkComboBoxText *cboxtext,
						     const gchar     *string);
static void    gtk_combo_box_text_tree_view_clicked(GtkTreeView *tree_view,
						    GtkComboBoxText *cboxtext);

static gint                  gtk_combo_box_text_signals[LAST_SIGNAL] = {0};
static GtkComboBoxClass     *gtk_combo_box_text_parent_class = NULL;

/* object class functions */

static void
gtk_combo_box_text_destroy (GtkObject *object)
{
	GtkComboBoxText  *cboxtext;
	
	g_return_if_fail (object != NULL);
	g_return_if_fail (GTK_IS_COMBO_BOX_TEXT (object));
	
	cboxtext = GTK_COMBO_BOX_TEXT (object);
	
	cboxtext->entry = NULL;
	cboxtext->scrolled_window = NULL;

	if (cboxtext->tree_view)
		gtk_widget_unref (cboxtext->tree_view);
	cboxtext->tree_view = NULL;

	if (cboxtext->cell)
		gtk_object_unref (GTK_OBJECT (cboxtext->cell));
	cboxtext->cell = NULL;

	if (cboxtext->model)
		gtk_object_unref (cboxtext->model);
	cboxtext->model = NULL;
	
	if (cboxtext->column)
		gtk_object_unref (cboxtext->column);
	cboxtext->column = NULL;

	GTK_OBJECT_CLASS (gtk_combo_box_text_parent_class)->destroy (object);
}

/* class functions */

static void
gtk_combo_box_text_class_init (GtkComboBoxTextClass *class)
{
	GtkObjectClass    *object_class;
	GtkWidgetClass    *widget_class;
	GtkContainerClass *container_class;
	
	g_return_if_fail (class != NULL);

	gtk_combo_box_text_parent_class = g_type_class_peek_parent (class);
	object_class = (GtkObjectClass *)class;
	widget_class = (GtkWidgetClass *)class;
	container_class = (GtkContainerClass *)class;

	object_class->destroy = gtk_combo_box_text_destroy;
	
	gtk_combo_box_text_signals[SELECT] =
		gtk_signal_new ( 
			"select",
			GTK_RUN_LAST,
			gtk_combo_box_text_get_type (),
			GTK_SIGNAL_OFFSET (GtkComboBoxTextClass, select),
			gtk_signal_default_marshaller,
			GTK_TYPE_NONE, 0);
	gtk_combo_box_text_signals[AUTOCOMPLETED] =
		gtk_signal_new (
			"autocompleted",
			GTK_RUN_LAST,
			gtk_combo_box_text_get_type (),
			GTK_SIGNAL_OFFSET (GtkComboBoxTextClass, select),
			gtk_signal_default_marshaller,
			GTK_TYPE_NONE, 0);
	
	gtk_object_class_add_signals (object_class,
				      gtk_combo_box_text_signals,
				      LAST_SIGNAL);
}

/* utility functions */

static void
gtk_combo_box_text_add_history (GtkComboBoxText    *cboxtext,
				const gchar        *string)
{
	gtk_combo_box_text_add_string (cboxtext, string);
}

static GSList *
gtk_combo_box_text_build_item_list (GtkTreeModel *model)
{
	GSList        *strings = NULL;
	GtkTreePath   *path;
	GtkTreeNode    node;
	
	g_return_val_if_fail (model != NULL, NULL);
	g_return_val_if_fail (GTK_IS_TREE_MODEL (model), NULL);
	
	path = gtk_tree_path_new_from_string ("0");
	node = gtk_tree_model_get_node (GTK_TREE_MODEL (model), path);
	g_free (path);
	
	g_return_val_if_fail (node != NULL, NULL);
	
	gtk_combo_box_text_add_to_item_list (&strings,
					     model,
					     node);
	
	while (gtk_tree_model_node_next (GTK_TREE_MODEL (model), &node))
		gtk_combo_box_text_add_to_item_list (&strings,
						     model, node);
	
	return strings;
}

static void
gtk_combo_box_text_add_to_item_list (GSList       **slist,
				     GtkTreeModel  *model,
				     GtkTreeNode    node)
{
	gchar    *string;
	
	string = g_strdup (gtk_combo_box_text_node_get_value (model, node));

	if (string)
		*slist = g_slist_append (*slist, string);
	
	return;
}

static void
gtk_combo_box_text_entry_changed (GtkComboBoxText *cboxtext)
{
	gchar *tmp = gtk_entry_get_text (GTK_ENTRY (cboxtext->entry));

	gtk_combo_box_text_select (cboxtext, tmp);
}

static gint
gtk_combo_box_text_entry_key_press (GtkComboBoxText *cboxtext,
				    GdkEventKey     *event,
				    GtkEntry        *entry)
{
	gboolean handled = FALSE;
	
	if (!cboxtext->autocomplete)
		return handled;
	
	if ((event->keyval == GDK_Tab || event->keyval == GDK_ISO_Left_Tab)
	    && !(event->state & (GDK_SHIFT_MASK | GDK_CONTROL_MASK
				 | GDK_MOD1_MASK))) {
		handled = TRUE;
		
		if (event->type == GDK_KEY_PRESS) {
			gchar *mstring;
			gchar *estring;

			mstring = gtk_combo_box_text_try_complete (cboxtext);
			estring = gtk_entry_get_text 
				(GTK_ENTRY (cboxtext->entry));
			
			if (!gtk_combo_box_text_match (cboxtext,
						       estring))
				if (mstring && strcmp (mstring, estring))
					gtk_entry_set_text 
						(GTK_ENTRY (cboxtext->entry),
						 mstring);
			
			g_free (mstring);

			gtk_signal_emit_stop_by_name 
				(GTK_OBJECT (cboxtext->entry),
				 "key-press-event");
		}
	} else if ((event->keyval == GDK_Return) 
		   || (event->keyval == GDK_KP_Enter)) {
		gchar *tmp;
		gchar *estring = gtk_entry_get_text
			(GTK_ENTRY (cboxtext->entry));
		
		handled = TRUE;
		
		if (estring && strlen (estring) > 1 
		    && event->type == GDK_KEY_PRESS) {
			if (gtk_combo_box_text_match (cboxtext, estring)) {
				tmp = gtk_entry_get_text 
					(GTK_ENTRY (cboxtext->entry));

				gtk_signal_emit 
					(GTK_OBJECT(cboxtext),
					 gtk_combo_box_text_signals
					 [AUTOCOMPLETED], tmp);
			} else {
				tmp = gtk_entry_get_text
					(GTK_ENTRY (cboxtext->entry));

				if (cboxtext->keep_history)
					gtk_combo_box_text_add_history 
						(cboxtext, tmp);
			}
		}
		
		gtk_signal_emit_stop_by_name (GTK_OBJECT (cboxtext->entry),
					      "key-press-event");
	}
	
	return handled;
}

static gchar *
gtk_combo_box_text_intersect (gint max_len, GSList *strings)
{
	gint   l = 0;
	gchar *completion;

	g_return_val_if_fail (max_len != 0, NULL);
	g_return_val_if_fail (strings != NULL, NULL);
	
	completion = g_new (gchar, max_len + 1);
	
	while (l < max_len) {
		gchar   *s = strings->data;
		GSList  *slist;

		s += l;
		completion[l] = *s;
		
		for (slist = strings->next; slist; slist = slist->next) {
			s = slist->data;
			s += l;
			if (completion[l] != *s)
				completion[l] = 0;
		}

		if (!completion[l])
			break;

		l++;
	}
	
	completion[l] = 0;
	
	return g_renew (gchar, completion, completion[0] ? l + 1 : 0);
}

static gboolean
gtk_combo_box_text_match (GtkComboBoxText *cboxtext,
			  const gchar     *string)
{
	gchar   *match = NULL;
	GSList  *s, *strings = NULL;

	if (strlen (string) < 1)
		return FALSE;

	strings = gtk_combo_box_text_build_item_list
		(GTK_TREE_MODEL (cboxtext->model));
	
	if (!strings)
		return FALSE;
	
	for (s = strings; s; s = s->next) {
		gchar *tmp = g_new (gchar, strlen(string));
		
		strncpy (tmp, s->data, strlen(string));
		
		if (!strcmp (tmp, string)) {
			if (!match)
				match = g_strdup (s->data);
			else {
				g_slist_free (strings);
				return FALSE;
			}
		}
	}

	g_slist_free (strings);

	if (!match)
		return FALSE;
	
	gtk_entry_set_text (GTK_ENTRY (cboxtext->entry), match);

	return TRUE;
}

static inline gchar *
gtk_combo_box_text_node_get_value (GtkTreeModel *model,
				   GtkTreeNode   node)
{
	GValue   value = {0,};
	
	gtk_tree_model_node_get_value (GTK_TREE_MODEL (model),
				       node, 0, &value);
	
	return g_value_get_string (&value);
}

static GtkTreeNode *
gtk_combo_box_text_node_new (GtkComboBoxText *cboxtext,
			     const gchar     *string)
{
	GValue        value = {0,};
	gchar        *str;
	GtkTreeNode  *node = gtk_tree_store_node_new ();
	
	g_return_val_if_fail (cboxtext != NULL, NULL);
	g_return_val_if_fail (GTK_IS_COMBO_BOX_TEXT (cboxtext), NULL);
	g_return_val_if_fail (string != NULL, NULL);

	g_value_init (&value, G_TYPE_STRING);
	str = g_strdup (string);
	g_value_set_string (&value, str);
	g_free (str);
	
	gtk_tree_store_node_set_cell (GTK_TREE_STORE (cboxtext->model),
				      node, 0, &value );
	g_value_unset (&value);
	
	return node;
}

static void
gtk_combo_box_text_tree_view_clicked (GtkTreeView     *tree_view,
				      GtkComboBoxText *cboxtext)
{
	GtkTreeSelection  *selection;

	g_return_if_fail (cboxtext != NULL);
	g_return_if_fail (GTK_IS_COMBO_BOX_TEXT (cboxtext));
	g_return_if_fail (tree_view != NULL);
	
	selection = gtk_tree_view_get_selection 
		(GTK_TREE_VIEW(cboxtext->tree_view));
	
	if (selection) {
		gchar       *string;
		GValue       value = {0,};
		GtkTreeNode  node;

		node = gtk_tree_selection_get_selected 
			(GTK_TREE_SELECTION (selection));
		
		if (!node)
			return;
		
		gtk_tree_model_node_get_value (GTK_TREE_MODEL(cboxtext->model),
					       node, 0, &value);
		string = g_strdup (g_value_get_string (&value));

		if (!string)
			return;
		
		gtk_entry_set_text (GTK_ENTRY (cboxtext->entry), string);
		g_free (string);

		gtk_combo_box_popup_hide (GTK_COMBO_BOX (cboxtext));
	}
}

/* header */

void
gtk_combo_box_text_add_string (GtkComboBoxText *cboxtext,
			       const gchar     *string)
{
	g_return_if_fail (cboxtext != NULL);
	g_return_if_fail (GTK_IS_COMBO_BOX_TEXT (cboxtext));
	g_return_if_fail (string != NULL);
	
	gtk_tree_store_node_append ( 
		GTK_TREE_STORE (cboxtext->model),
		NULL,
		gtk_combo_box_text_node_new (cboxtext, string));
}

void
gtk_combo_box_text_construct (GtkComboBoxText *cboxtext)
{
	GtkObject    *selection;

	cboxtext->keep_history = cboxtext->autocomplete = FALSE;

	cboxtext->entry = gtk_entry_new ();
	gtk_widget_ref (cboxtext->entry);
	gtk_signal_connect_object ( 
		GTK_OBJECT (cboxtext->entry), "changed",
		GTK_SIGNAL_FUNC (gtk_combo_box_text_entry_changed),
		GTK_OBJECT (cboxtext));
	gtk_signal_connect_object ( 
		GTK_OBJECT (cboxtext->entry), "key_press_event",
		GTK_SIGNAL_FUNC (gtk_combo_box_text_entry_key_press),
		GTK_OBJECT (cboxtext));
	gtk_widget_show (cboxtext->entry);

	cboxtext->scrolled_window = 
		gtk_scrolled_window_new (NULL, NULL);
	gtk_widget_set_usize (cboxtext->scrolled_window,
			      100, 75); /* FIXME: this is ugly */
	gtk_scrolled_window_set_policy ( 
		GTK_SCROLLED_WINDOW (cboxtext->scrolled_window),
		GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	
	cboxtext->model = gtk_tree_store_new_with_values (1,
							  G_TYPE_STRING);
	cboxtext->tree_view = gtk_tree_view_new_with_model ( 
		GTK_TREE_MODEL (cboxtext->model));
	gtk_widget_ref (cboxtext->tree_view);
	selection = GTK_OBJECT (gtk_tree_view_get_selection (
		GTK_TREE_VIEW (cboxtext->tree_view )));
	gtk_tree_selection_set_type (GTK_TREE_SELECTION (selection),
				     GTK_TREE_SELECTION_SINGLE );
	gtk_signal_connect (
		GTK_OBJECT (selection), "selection_changed",
		GTK_SIGNAL_FUNC(gtk_combo_box_text_tree_view_clicked),
		cboxtext);
	gtk_container_add (GTK_CONTAINER (cboxtext->scrolled_window),
			   cboxtext->tree_view);
	
	cboxtext->cell = gtk_cell_renderer_text_new ();
	cboxtext->column = gtk_tree_view_column_new_with_attributes (
		"nodes", cboxtext->cell, "text", 0, NULL);
	gtk_tree_view_set_headers_visible ( 
		GTK_TREE_VIEW (cboxtext->tree_view), FALSE);
	/* FIXME: write some code, so the tree will be auto sorted */
	gtk_tree_view_append_column (
		GTK_TREE_VIEW (cboxtext->tree_view),
		GTK_TREE_VIEW_COLUMN (cboxtext->column));

	gtk_widget_show_all (cboxtext->tree_view);
	
	gtk_combo_box_construct (GTK_COMBO_BOX (cboxtext),
				 cboxtext->entry,
				 cboxtext->scrolled_window);
}

GSList *
gtk_combo_box_text_get_strings (GtkComboBoxText *cboxtext)
{
	g_return_val_if_fail (cboxtext != NULL, NULL);
	g_return_val_if_fail (GTK_IS_COMBO_BOX_TEXT (cboxtext), NULL);

	return gtk_combo_box_text_build_item_list 
		(GTK_TREE_MODEL(cboxtext->model));
}

GtkType
gtk_combo_box_text_get_type (void)
{
	static GtkType   type = 0;
	
	if (!type) {
		static const GTypeInfo info = {
			sizeof (GtkComboBoxTextClass),
			NULL, /* GBaseInitFunc */
			NULL, /* GBaseFinalizeFunc */
			(GClassInitFunc) gtk_combo_box_text_class_init,
			NULL,
			NULL, /* class data */
			sizeof (GtkComboBoxText),
			0, NULL
		};

		type = g_type_register_static (gtk_combo_box_get_type (),
					       "GtkComboBoxText",
					       &info );
	}
	
	return type;
}

GtkWidget *
gtk_combo_box_text_new (void)
{
	GtkComboBoxText  *cboxtext;

	cboxtext = gtk_type_new (gtk_combo_box_text_get_type ());
	gtk_combo_box_text_construct (cboxtext);
	
	return GTK_WIDGET (cboxtext);
}

void
gtk_combo_box_text_remove_string (GtkComboBoxText *cboxtext,
				  const gchar     *string)
{
	GtkTreePath     *path;
	GtkTreeNode      node;

	g_return_if_fail (cboxtext != NULL);
	g_return_if_fail (GTK_IS_COMBO_BOX_TEXT (cboxtext));
	g_return_if_fail (string != NULL);

	path = gtk_tree_path_new_from_string ("0");
	node = gtk_tree_model_get_node (GTK_TREE_MODEL (cboxtext->model),
					path);
	g_free (path);
	
	g_return_if_fail (node != NULL);
	
	if (!strcmp (string, gtk_combo_box_text_node_get_value 
		     (GTK_TREE_MODEL (cboxtext->model), node)))
		gtk_tree_store_node_remove (GTK_TREE_STORE (cboxtext->model),
					    node);
	
	while (gtk_tree_model_node_next (GTK_TREE_MODEL (cboxtext->model),
					 &node)) {
		if (!strcmp (string, gtk_combo_box_text_node_get_value
			     (GTK_TREE_MODEL (cboxtext->model), node)))
			gtk_tree_store_node_remove 
				(GTK_TREE_STORE (cboxtext->model),
				 node);
	}
}

void
gtk_combo_box_text_remove_matches (GtkComboBoxText *cboxtext,
				   const gchar     *pattern)
{
	GtkTreePath     *path;
	GtkTreeNode      node;
	GtkPatternSpec   pspec;

	g_return_if_fail (cboxtext != NULL);
	g_return_if_fail (GTK_IS_COMBO_BOX_TEXT (cboxtext));
	
	if (!pattern)
		pattern = "*";
	
	gtk_pattern_spec_init (&pspec, pattern);	
	
	path = gtk_tree_path_new_from_string ("0");
	node = gtk_tree_model_get_node (GTK_TREE_MODEL (cboxtext->model),
					path);
	g_free (path);
	
	g_return_if_fail (node != NULL);
	
	if (gtk_pattern_match_string (&pspec, gtk_combo_box_text_node_get_value (GTK_TREE_MODEL (cboxtext->model), node)))
		gtk_tree_store_node_remove
			(GTK_TREE_STORE (cboxtext->model), node);
	
	while (gtk_tree_model_node_next (GTK_TREE_MODEL (cboxtext->model),
					 &node))
		if (gtk_pattern_match_string (&pspec, gtk_combo_box_text_node_get_value (GTK_TREE_MODEL (cboxtext->model), node)))
			gtk_tree_store_node_remove 
				(GTK_TREE_STORE (cboxtext->model), node);
	
	gtk_pattern_spec_free_segs (&pspec);
}

void
gtk_combo_box_text_set_autocomplete (GtkComboBoxText *cboxtext,
				     gboolean autocomplete)
{
	g_return_if_fail (cboxtext != NULL);
	g_return_if_fail (GTK_IS_COMBO_BOX_TEXT (cboxtext));
	
	cboxtext->autocomplete = autocomplete;
}

void
gtk_combo_box_text_set_keep_history (GtkComboBoxText *cboxtext,
				     gboolean keep_history)
{
	g_return_if_fail (cboxtext != NULL);
	g_return_if_fail (GTK_IS_COMBO_BOX_TEXT (cboxtext));
	
	cboxtext->keep_history = keep_history;
}

void
gtk_combo_box_text_select (GtkComboBoxText *cboxtext,
			   const gchar     *string)
{
	g_return_if_fail (cboxtext != NULL);
	g_return_if_fail (GTK_IS_COMBO_BOX_TEXT (cboxtext));
	g_return_if_fail (string != NULL);
	
	gtk_signal_emit (GTK_OBJECT (cboxtext),
			 gtk_combo_box_text_signals[SELECT],
			 string);
}

gchar *
gtk_combo_box_text_try_complete (GtkComboBoxText *cboxtext)
{
	gint             max_len = 0;
	gchar           *completion;
	GSList          *s;
	GSList          *strings = NULL;
	
	g_return_val_if_fail (cboxtext != NULL, NULL);
	g_return_val_if_fail (GTK_IS_COMBO_BOX_TEXT (cboxtext), NULL);
	
	strings = gtk_combo_box_text_build_item_list 
		(GTK_TREE_MODEL(cboxtext->model));
	
	if (!strings)
		return NULL;
	
	for (s = strings; s; s = s->next)
		max_len = MAX (max_len, strlen (s->data));

	completion = gtk_combo_box_text_intersect (max_len, strings);
	g_slist_free (strings);
	
	return completion;
}
